/*
 * unicode.c - Copyright (c) 2004 GeekNET. All rights reserved.
 * 
 * postgresql support functions for handling UTF8 encoded character streams.
 *
*/
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "utils/builtins.h"
#include "mb/pg_wchar.h"
#include "unicode.h"

typedef struct {
    UTF32       character;
    UTF32       upper;
    UTF32       lower;
    char        *category;
} unicode_map;

typedef struct {
    UTF32       character;
    UTF32       upper[4];
    UTF32       lower[4];
} unicode_specialmap;

#include "unicodedata.h"

/*
 *  verify that the UTF8 character pointed to by source with length len is valid UTF8
 */

bool isLegalUTF8(const UTF8 *source, int len) {
        UTF8 a;
        const UTF8 *srcptr = source+len;
        if(!source || (pg_utf_mblen(source) != len)) return false;
        switch (len) {
            default: return false;
            /* Everything else falls through when "true"... */
            case 6: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
            case 5: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
            case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
            case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
            case 2: if ((a = (*--srcptr)) > 0xBF) return false;
            switch (*source) {
                    /* no fall-through in this inner switch */
                    case 0xE0: if (a < 0xA0) return false; break;
                    case 0xF0: if (a < 0x90) return false; break;
                    case 0xF4: if (a > 0x8F) return false; break;
                    default:  if (a < 0x80) return false;
            }
            case 1: if (*source >= 0x80 && *source < 0xC2) return false;
            if (*source > 0xFD) return false;
        }
        return true;
}


/*
 *  calculate the byte length of the UTF8 character pointed to by s
 */

static int utf8_len(const UTF8 *s) {
        int	len = 1;

        if 	((*s & 0x80) == 0)	len = 1;
        else if	((*s & 0xe0) == 0xc0)	len = 2;
        else if	((*s & 0xf0) == 0xe0)	len = 3;
        else if	((*s & 0xf8) == 0xf0)	len = 4;
        else if	((*s & 0xfc) == 0xf8)	len = 5;
        else if	((*s & 0xfe) == 0xfc)	len = 6;
        return	(len);
}

/*
 *  Encode a single UTF8 character pointed to by the utf8 parameter as a UTF32 value
 */

static UTF32 utf8_to_wchar(UTF8 *utf8) {
    const UTF32 offsetsFromUTF8[6] = { 0x00000000UL, 0x00003080UL, 0x000E2080UL, 0x03C82080UL, 0xFA082080UL, 0x82082080UL };
    unsigned int utflen = utf8_len(utf8);
    UTF32 ch = 0;
    
    switch (utflen) {
        case 6: ch += *utf8++; ch <<= 6;
        case 5: ch += *utf8++; ch <<= 6;
        case 4: ch += *utf8++; ch <<= 6;
        case 3: ch += *utf8++; ch <<= 6;
        case 2: ch += *utf8++; ch <<= 6;
        case 1: ch += *utf8++;
    }
    ch -= offsetsFromUTF8[utflen - 1];
    return ch;
}

/*
 *  Encode the single UTF32 character in the utf32 parameter as a UTF8 string
 *  caller is responsible for freing memory allocated for the string.
 */

static UTF8 *wchar_to_utf8(UTF32 utf32) {
    static const UTF8 firstByteMark[6] = { 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC };
    unsigned short bytesToWrite = 0;
    const UTF32 byteMask = 0xBF;
    const UTF32 byteMark = 0x80;

    if		(utf32 < (UTF32)0x00000080)	{	bytesToWrite = 1;
    } else if	(utf32 < (UTF32)0x00000800)	{	bytesToWrite = 2;
    } else if	(utf32 < (UTF32)0x00010000)	{	bytesToWrite = 3;
    } else if	(utf32 < (UTF32)0x00200000)	{	bytesToWrite = 4;
    } else if	(utf32 < (UTF32)0x04000000)	{	bytesToWrite = 5;
    } else					{	bytesToWrite = 6;
    }
	
    UTF8 *result = palloc0(sizeof(UTF8) * (bytesToWrite + 1));
    result += bytesToWrite;

    switch (bytesToWrite) { /* note: everything falls through. */
        case 6: *--result = (UTF8)((utf32 | byteMark) & byteMask); utf32 >>= 6;
        case 5: *--result = (UTF8)((utf32 | byteMark) & byteMask); utf32 >>= 6;
        case 4: *--result = (UTF8)((utf32 | byteMark) & byteMask); utf32 >>= 6;
        case 3: *--result = (UTF8)((utf32 | byteMark) & byteMask); utf32 >>= 6;
        case 2: *--result = (UTF8)((utf32 | byteMark) & byteMask); utf32 >>= 6;
        case 1: *--result = (UTF8) (utf32 | firstByteMark[bytesToWrite - 1]);
    }
    return result;
}

/*
 * unicode lookup table btree support functions
 */

static int utf_comp(const void *m1, const void *m2) {
    unicode_map *mi1 = (unicode_map *) m1;
    unicode_map *mi2 = (unicode_map *) m2;
    if( mi1->character < mi2->character ) return -1;
    if( mi1->character == mi2->character ) return 0;
    if( mi1->character > mi2->character ) return 1;
}

static unicode_map *utf_casemap_findkey(UTF32 search) {
    unicode_map key, *res;

    static int unsorted = 1;
    static unicode_map sorted[(sizeof(utf_casemap)/sizeof(utf_casemap[0]))];
    if(unsorted) {
        memcpy(sorted,utf_casemap,sizeof(utf_casemap));
        qsort(sorted, (sizeof(utf_casemap)/sizeof(utf_casemap[0])), sizeof(unicode_map), utf_comp);
        unsorted = 0;
    }

    key.character = search;
    res = bsearch(&key, sorted, sizeof(sorted)/sizeof(sorted[0]), sizeof(unicode_map), utf_comp);
    return res;
}

static int utf_specialcomp(const void *m1, const void *m2) {
    unicode_specialmap *mi1 = (unicode_specialmap *) m1;
    unicode_specialmap *mi2 = (unicode_specialmap *) m2;
    if( mi1->character < mi2->character ) return -1;
    if( mi1->character == mi2->character ) return 0;
    if( mi1->character > mi2->character ) return 1;
}

static unicode_specialmap *utf_specialmap_findkey(UTF32 search) {
    unicode_specialmap key, *res;

    static int unsorted = 1;
    static unicode_specialmap sorted[(sizeof(utf_specialmap)/sizeof(utf_specialmap[0]))];
    if(unsorted) {
        memcpy(sorted,utf_specialmap,sizeof(utf_specialmap));
        qsort(sorted, (sizeof(utf_specialmap)/sizeof(utf_specialmap[0])), sizeof(unicode_specialmap), utf_specialcomp);
        unsorted = 0;
    }

    key.character = search;
    res = bsearch(&key, sorted, sizeof(sorted)/sizeof(sorted[0]), sizeof(unicode_specialmap), utf_specialcomp);
    return res;
}

/*
 * convert a utf8 string to either upper or lower case as specified.
 */

static UTF8 *str_upper_lower(UTF8 *utf8, bool upper) {
    UTF8 *begin				= utf8;
    UTF8 *end				= utf8 + strlen(utf8);
    UTF8 *result			= palloc0(sizeof(UTF8));
    UTF8 *utf8char			= NULL;
    unicode_map *finder			= NULL;
    unicode_specialmap *specialfinder	= NULL;
    UTF32 utfchar;
    int utflen,i;
    
    while(begin < end) {
	if(!isLegalUTF8(begin,utf8_len(begin))) {
	    elog(WARNING,"Invalid UNICODE byte sequence detected near character %c. Skipping until reach of valid sequence",*begin);
	    while((++begin) < end) {
		if(isLegalUTF8(begin,utf8_len(begin)))
		    break;
	    }
	}
	utfchar = utf8_to_wchar(begin);
	specialfinder = utf_specialmap_findkey(utfchar);
	if(specialfinder) {
	    i = 0;
	    while(upper ? specialfinder->upper[i] : specialfinder->lower[i]) {
	        utf8char = wchar_to_utf8(upper ? specialfinder->upper[i] : specialfinder->lower[i]);
		    utflen = utf8_len(utf8char);
		    result = (UTF8*) repalloc(result,(sizeof(UTF8) * (strlen(result) + utflen + 1)));
		    strncat(result,utf8char,utflen);
		pfree(utf8char);
		i++;
	    }
	} else {
	    finder = utf_casemap_findkey(utfchar);
	    utf8char = wchar_to_utf8(finder ? (upper ? finder->upper : finder->lower) : utfchar);
		utflen = utf8_len(utf8char);
		result = (UTF8*) repalloc(result,(sizeof(UTF8) * (strlen(result) + utflen + 1)));
		strncat(result,utf8char,utflen);
	    pfree(utf8char);
	}

	begin += utf8_len(begin);
    }
    return result;
}

UTF8 *str_toupper(UTF8 *utf8) {
    return utf8 ? str_upper_lower(utf8,true) : NULL;
}

UTF8 *str_tolower(UTF8 *utf8) {
    return utf8 ? str_upper_lower(utf8,false) : NULL;
}
